@@ -198,25 +198,6 @@ gp11_slot_class_init (GP11SlotClass *klass)
 	g_type_class_add_private (gobject_class, sizeof (GP11SlotPrivate));
 }
 
-/* ----------------------------------------------------------------------------
- * INTERNAL AUTHENTICATION
- */
-
-gboolean
-_gp11_slot_is_protected_auth_path (GP11Slot *self)
-{
-	GP11TokenInfo *info;
-	gboolean ret;
-
-	g_assert (GP11_IS_SLOT (self));
-
-	info = gp11_slot_get_token_info (self);
-	ret = (info && info->flags & CKF_PROTECTED_AUTHENTICATION_PATH);
-	gp11_token_info_free (info);
-	
-	return ret;
-}
-
 /* ----------------------------------------------------------------------------
  * PUBLIC 
  */
@@ -269,6 +250,101 @@ gp11_mechanism_info_free (GP11MechanismInfo *mech_info)
 	g_free (mech_info);
 }
 
+/**
+ * gp11_mechanisms_check:
+ * @mechanisms: A list of mechanisms, perhaps retrieved from gp11_slot_get_mechanisms().
+ * 
+ * Check whether all the mechanism types are in the list.
+ * 
+ * The arguments should be a list of CKM_XXX mechanism types. The last argument
+ * should be GP11_INVALID. 
+ * 
+ * Return value: Whether the mechanism is in the list or not.
+ **/
+gboolean
+gp11_mechanisms_check (GP11Mechanisms *mechanisms, ...)
+{
+	gboolean found = TRUE;
+	va_list va;
+	gulong mech;
+	gsize i;
+	
+	g_return_val_if_fail (mechanisms, FALSE);
+	
+	va_start (va, mechanisms);
+	for (;;) {
+		mech = va_arg (va, gulong);
+		if (mech == GP11_INVALID)
+			break;
+		
+		found = FALSE;
+		for (i = 0; i < gp11_mechanisms_length (mechanisms); ++i) {
+			if (gp11_mechanisms_at (mechanisms, i) == mech) {
+				found = TRUE;
+				break;
+			}
+		}
+		
+		if (found == FALSE)
+			break;
+		
+	}
+	va_end (va);
+	
+	return found;
+}
+
+/**
+ * gp11_slot_equal:
+ * @slot1: A pointer to the first GP11Slot
+ * @slot2: A pointer to the second GP11Slot
+ * 
+ * Checks equality of two slots. Two GP11Slot objects can point to the same 
+ * underlying PKCS#11 slot.
+ * 
+ * Return value: TRUE if slot1 and slot2 are equal. FALSE if either is not a GP11Slot.
+ **/
+gboolean
+gp11_slot_equal (gconstpointer slot1, gconstpointer slot2)
+{
+	GP11SlotData *data1, *data2;
+
+	if (slot1 == slot2)
+		return TRUE;
+	if (!GP11_IS_SLOT (slot1) || !GP11_IS_SLOT (slot2))
+		return FALSE;
+	
+	data1 = GP11_SLOT_GET_DATA (slot1);
+	data2 = GP11_SLOT_GET_DATA (slot2);
+	
+	return data1->handle == data2->handle && 
+	       gp11_module_equal (data1->module, data2->module);
+}
+
+/**
+ * gp11_slot_hash:
+ * @slot: A pointer to a GP11Slot
+ * 
+ * Create a hash value for the GP11Slot. 
+ * 
+ * This function is intended for easily hashing a GP11Slot to add to 
+ * a GHashTable or similar data structure.
+ * 
+ * Return value: An integer that can be used as a hash value, or 0 if invalid.
+ **/
+guint
+gp11_slot_hash (gconstpointer slot)
+{
+	GP11SlotData *data;
+	
+	g_return_val_if_fail (GP11_IS_SLOT (slot), 0);
+
+	data = GP11_SLOT_GET_DATA (slot);
+	
+	return _gp11_ulong_hash (&data->handle) ^
+	       gp11_module_hash (data->module);
+}
+
 /**
  * gp11_slot_get_handle:
  * @self: The slot to get the handle of.
@@ -444,7 +520,7 @@ gp11_slot_get_mechanisms (GP11Slot *self)
 	CK_SLOT_ID handle = (CK_SLOT_ID)-1;
 	CK_FUNCTION_LIST_PTR funcs;
 	GP11Module *module = NULL;
-	CK_MECHANISM_TYPE_PTR mech_list;
+	CK_MECHANISM_TYPE_PTR mech_list = NULL;
 	CK_ULONG count, i;
 	GP11Mechanisms *result;
 	CK_RV rv;
@@ -532,6 +608,36 @@ gp11_slot_get_mechanism_info (GP11Slot *self, gulong mech_type)
 	return mechinfo;
 }
 
+gboolean
+gp11_slot_has_flags (GP11Slot *self, gulong flags)
+{
+	CK_FUNCTION_LIST_PTR funcs;
+	GP11Module *module = NULL;
+	CK_TOKEN_INFO info;
+	CK_SLOT_ID handle;
+	CK_RV rv;
+	
+	g_return_val_if_fail (GP11_IS_SLOT (self), FALSE);
+
+	g_object_get (self, "module", &module, "handle", &handle, NULL);
+	g_return_val_if_fail (GP11_IS_MODULE (module), FALSE);
+	
+	funcs = gp11_module_get_functions (module);
+	g_return_val_if_fail (funcs, FALSE);
+	
+	memset (&info, 0, sizeof (info));
+	rv = (funcs->C_GetTokenInfo) (handle, &info);
+	
+	g_object_unref (module);
+	
+	if (rv != CKR_OK) {
+		g_warning ("couldn't get slot info: %s", gp11_message_from_rv (rv));
+		return FALSE;
+	}
+	
+	return (info.flags & flags) != 0;
+}
+
 #if UNIMPLEMENTED
 
 typedef struct InitToken {
@@ -639,7 +745,7 @@ complete_open_session (OpenSession *args, CK_RV result)
 		
 		ret = _gp11_module_fire_authenticate_slot (module, args->slot, NULL, &args->password);
 		
-		/* Call is not complete */
+		/* If authenticate returns TRUE then call is not complete */
 		ret = !ret;
 	}
 
